.onLoad <- function(lib, pkg)
{
    packageStartupMessage(paste("LSB.CMM package V1.0, by Li-Ping Liu")) 
}


# Normalize each row (logrithm values) to be logrithm of a multinomial probability. 

#input
# logProb: each row is logrithm of values proportional to some probability 

#return value:
# logProb: each row is logrithm of multinomial probability 

logNorm <- function(logProb)
{
    mv <- apply(logProb, 1, max)
    logProb <- t(scale(t(logProb), scale=FALSE, center=mv))
    prob <- exp(logProb)
    norm <- apply(prob, 1, sum)
    lognorm <- matrix(log(norm), ncol=1)
    logProb <- logProb - lognorm[, rep(1, ncol(logProb))]

    attr(logProb, "scaled:center") <- NULL
    return(logProb)
}

# stick breaking process with probability P

# input:
# P: each row are Bernoulli probabilities of choosing each component

# return value:
# phi: each row are multinomial probabilities of choosing each component

stickBreak <- function(P)
{
    phi <- P 
    temp <- 1 - P 

    for (k in 2:ncol(P)) 
    {
        phi[, k] <- temp[, k - 1] * P[, k]
        temp[, k] <- temp[, k - 1] * (1 - P[, k])
    }
    return(phi)
}

# stick breaking process with with logrithm values.  

#args:
# X: covariates of the data 
# W: weights for logistic function

#return value
# logTheta: each row is logrithm of multinomial probabilities of choosing each component

logStickBreak <- function(X, W)
{
    link <- X %*% W
    logV1 <- - log(1 + exp(-link))
    logV1[-link > 30] <- link[-link > 30]
    logV0 <- - log(1 + exp(link))
    logV0[link > 30] <- - link[link > 30] 

    accum <- logV0
    accum[, 1] <- 0

    for (k in 2:ncol(logV1)) 
    {
        accum[, k] <- accum[, k - 1] + logV0[, k - 1]
    }
    logPhi <- logV1 + accum 
    logPhi <- logNorm(logPhi)
    return(logPhi)
}

# the objective value of M step
fn <- function(Wk, X, phik, phis, sigma2)
{
    D <- length(Wk)
    Wk <- matrix(Wk, nrow=D, ncol=1)

    obj <- 0.5 * sum(Wk[2:D] * Wk[2:D]) / sigma2
  
    negLink <- - X %*% Wk
    logPos <- log(1 + exp(negLink))
    logPos[negLink > 30] <- negLink[negLink > 30]

    posLink <- - negLink
    logNeg <- log(1 + exp(posLink))
    logNeg[posLink > 30] <- posLink[posLink > 30]

    obj <- obj + sum((phik * logPos) + phis * log(1 + logNeg))

    return(obj)
}

# the gradient of M step
gr <- function(Wk, X, phik, phis, sigma2)
{
    D <- length(Wk)
    Wk <- matrix(Wk, nrow=D, ncol=1)

    gradient <- rep(0, D)
    gradient[2:D] <- Wk[2:D] / sigma2

    partn <- - phik / (1 + exp(X %*% Wk)) + phis / (1 + exp(- X %*% Wk))

    gradient <- gradient + t(X) %*% partn 

    return(gradient)

}

ovaCode <- function(L, K)
{
    code <- matrix(0, nrow=L, ncol=K) 
    cls <- (0 : (K - 1)) %% L + 1
    for (k in 1:K)
    {
        code[cls[k], k] <- 1
    }
    code <- code * 10 + 1
    return(list(ealpha=code, cls=cls))
}

# calculate expectation of log theta and theta is from dirichlet distribution with alpha as parameter
# input:
# ealpha: parameter of the Dirichlet distribution

# return value:
# mean.log: expected value of log(\theta_i) for each i

ELogTheta <- function(ealpha)
{
    mean.log <- ealpha
    for (k in 1:ncol(ealpha))
    {
        x <- rdirichlet(10000, ealpha[, k])
        mean.log[, k] <- apply(log(x), 2, mean)
        if (any(mean.log == -Inf))
        {
            stop("The expectation becomes minus infinity")
        }
    }
    return(mean.log)
}

#fit a conditional mixture model
#input: 
# X: fitting covariates, with each row as an instance. 
# Y: target indicator matrix, with each row as 0-1 indicators of possible classes. One class is correct label.  
# sigma2: variance (\sigma^2) of the Gaussian prior over the weight matrix W. Play the role of C. 
# K: maximum number of components 
# alpha: prior of the Dirichlet process

#return value:
# model: the conditional mixture model

lsbcmm.fit <- function(X, Y, sigma2 = 1, K = (5 * ncol(Y)), alpha = 0.05, marg.theta = TRUE, quiet = FALSE)
{
    X <- data.matrix(X)
    N <- nrow(X) 
    X <- cbind(rep(1, N), X)
    names(X)[1] <- "(intercept)"
    D <- ncol(X)
    
    L <- ncol(Y)
    if (is.data.frame(Y))
    {
        Y <- data.matrix(Y)
    }
    indY <- Y
    labseq <- seq_len(L)
    Y <- lapply(seq_len(nrow(Y)), function(i) labseq[as.logical(Y[i, ])]) 

    if (!quiet)
    {
        print(paste(L, "classes"))
        print(paste("Using ", K, " components", sep=""))
    }

    model <- list()
    model$likelihood <- -Inf 
    W <- matrix(0, nrow = D, ncol = K)
    W[1, ] <- 0
    W[1, K] <- Inf 

    control <- list()
    control$fnscale <- 1
    control$maxit <- 100
    control$abstol <- 1e-8

    n.start <- 1
    for (trial in 1:n.start)
    {
        code <- ovaCode(L, K)
        ealpha <- code$ealpha

        # in the case theta is not marginalized, theta is initialized as normalized ealpha instead of (ealpha - 1)
        theta <- scale(ealpha, scale=colSums(ealpha), center=FALSE)
        attr(theta,"scaled:scale") <- NULL

        logPhi <- matrix(0, nrow=N, ncol=K)

        n.iter <- 100
        likelihood <- rep(0, n.iter) 
        for (iter in 1:n.iter)
        {
            #E step  
            if (marg.theta)
            {
                elog <- ELogTheta(ealpha) 
            } else 
            {
                elog <- log(theta)
                elog[is.nan(elog)] <- -10^9
            }

            eVar <- .Call("DecompE", alpha   = as.double(alpha), 
                                     logPhi  = as.double(t(logPhi)), 
                                     elog    = as.double(t(elog)), 
                                     label   = Y,  
                                     K       = as.integer(K),
                                     L       = as.integer(L),
                                     PACKAGE = "LSB.CMM")
            	                        

            names(eVar) <- c("count", "ephi", "entropy", "logpy")
            count <- matrix(eVar$count, nrow=L, ncol=K, byrow=TRUE)
            ealpha <- count + alpha
            ephi <- matrix(eVar$ephi, nrow=N, ncol=K, byrow=TRUE)
            entropy <- eVar$entropy
            logpy <- eVar$logpy

            if (marg.theta)
            {
                theta <- scale(ealpha, scale=colSums(ealpha), center=FALSE)
                attr(theta,"scaled:scale") <- NULL
            } else 
            {
                # theta is the mode of the Dirichlet with ealpha. If alpha < 1, 
                # then the mode is not accurate.
                theta <- ealpha - 1
                theta[theta < 0] <- 1e-9 
                theta <- scale(theta, scale=colSums(theta), center=FALSE)
                attr(theta,"scaled:scale") <- NULL
            }            

            #M step
            for (k in 1:(K - 1))
            {
                phik <- ephi[, k] 
                phis <- apply(matrix(ephi[, (k + 1):K], nrow=N), 1, sum) 
                solution <- optim(par=W[, k], fn=fn, gr=gr, X=X, phik=phik, phis=phis, sigma2=sigma2,
                                  method = "BFGS", control=control, hessian = FALSE)
                W[, k] <- solution$par
            }

            logPhi <- logStickBreak(X, W) 

            # calculate the lower bound to decide when to stop 
            if (marg.theta) 
            {
                term1 <- sum(ephi * logPhi)
                term3 <- logpy
                term4 <- entropy

                #term 2 and term5 are combined to be the KL divergence between two Dir dist. with \hat{alpha} and alpha
                elog <- ELogTheta(ealpha) 
                term5 <- sum(lgamma(colSums(ealpha))) - sum(lgamma(ealpha)) - K * lgamma(alpha * L) + K * (lgamma(alpha) * L) 
                         + sum((ealpha - alpha) * elog)

                term6 <- -0.5 * sum(W[2:D, 1:(K - 1)] * W[2:D, 1:(K - 1)]) / sigma2 

                likelihood[iter] <- term1 + term3 + term4 - term5 + term6 

                if (!quiet)
                    print(paste("Lower bound becomes", likelihood[iter]))

            } else
            {
                # probability of W with Gaussian prior
                term1 <- -0.5 * sum(W[2:D, 1:(K - 1)] * W[2:D, 1:(K - 1)]) / sigma2 

                # probability of \theta with Dirichlet prior
                term2 <- (alpha - 1) * sum(log(theta)) 

                term3 <- sum(ephi * logPhi) + sum(count * log(theta))    # Expectation of log p(y, z | X, W, \theta)
                term4 <- entropy                                          # Entropy term 
                
                likelihood[iter] <- term1 + term2 + term3 + term4
                if (!quiet)
                    print(paste("Lower bound becomes", likelihood[iter]))
            }

            if (iter > 10)
            {
                curLike <- mean(likelihood[(iter - 2) : iter])
                lastLike <- mean(likelihood[(iter - 5) : (iter - 3)])

                if ((curLike - lastLike) < 1e-4 * abs(curLike))
                {
                    break;
                }
            } 
        }

        if (model$likelihood < likelihood[n.iter])
        {
            model$W <- W
            model$theta <- theta 
            model$likelihood <- likelihood[n.iter]
        }
    }

    class(model) <- "LSBCMM"
    return(model)
}

#predict true labels of testing data 
#input:
# object: the conditional mixture model
# testX: the data with each row as an instance 
#output:
# the probability of being each class

predict.LSBCMM <- function(object, testX, ...) 
{
    testX <- data.matrix(testX)
    testX <- cbind(rep(1, nrow(testX)), testX) 
    P <- 1 / (1 + exp(- testX %*% object$W))
    phi <- stickBreak(P)

    prob <- phi %*% t(object$theta)
    return(prob)
}


