#! /usr/bin/env Rscript

# generate a tri-cluster classification problem with two classes.
# Centers of three clusters are on the x-axis. The middle cluster
# belongs to negative class and the other two are positive, so it
# islinearly inseperable. 
# input
#     var: variance of each Gaussian cluster
# output
#     X: covariates
#     Y: labels as a matrix of binary indicators
#     label: labels as a vector 

triClust <- function(var = 0.25)
{
    n.train <- 300
    X <- matrix(rnorm(n.train * 2, mean=0, sd=var), ncol=2)
    Y <- matrix(0, nrow=nrow(X), ncol=2)
    label <- matrix(0, nrow=nrow(X), ncol=1)
    
    posInd1 <- 1:(n.train / 3)
    posInd2 <- (2 * n.train / 3 + 1):n.train
    negInd <- (n.train / 3 + 1):(2 * n.train / 3)
    
    X[posInd1, 1] <- X[posInd1, 1] - 1
    Y[posInd1, 1] <- 1
    label[posInd1] <- 1
    
    X[posInd2, 1] <- X[posInd2, 1] + 1
    Y[posInd2, 1] <- 1
    label[posInd2] <- 1
    
    X[negInd, 1] <- X[negInd, 1] + 0
    Y[negInd, 2] <- 1
    label[negInd] <- 2
   
    return(list(X=X, Y=Y, label=label))
}

# generate a quad-cluster classification problem with two classes.
# Cluster centers are at the four corners of a square and the class 
# assignment makes it an XOR problem and linearly inseperable. 
# input
#     var: variance of each Gaussian cluster
# output
#     X: covariates
#     Y: labels as a matrix of binary indicators
#     label: labels as a vector 

quadClust <- function(var = 0.25)
{
    n.train <- 400
    X <- matrix(rnorm(n.train * 2, mean=0, sd=var), ncol=2)
    Y <- matrix(0, nrow=nrow(X), ncol=2)
    label <- matrix(0, nrow=nrow(X), ncol=1)
    
    ind1 <- (0 * n.train / 4 + 1):(1 * n.train / 4)
    ind2 <- (1 * n.train / 4 + 1):(2 * n.train / 4)
    ind3 <- (2 * n.train / 4 + 1):(3 * n.train / 4)
    ind4 <- (3 * n.train / 4 + 1):(4 * n.train / 4)
   
    center1 <- matrix(c(0.75, 0.75), ncol=2) 
    center2 <- matrix(c(-0.75, 0.75), ncol=2) 
    center3 <- matrix(c(-0.75, -0.75), ncol=2) 
    center4 <- matrix(c(0.75, -0.75), ncol=2) 
    
    X[ind1, ] <- X[ind1, ] + center1[rep(1, n.train / 4), ] 
    Y[ind1, 1] <- 1
    label[ind1] <- 1

    X[ind2, ] <- X[ind2, ] + center2[rep(1, n.train / 4), ] 
    Y[ind2, 2] <- 1
    label[ind2] <- 2

    X[ind3, ] <- X[ind3, ] + center3[rep(1, n.train / 4), ] 
    Y[ind3, 1] <- 1
    label[ind3] <- 1

    X[ind4, ] <- X[ind4, ] + center4[rep(1, n.train / 4), ] 
    Y[ind4, 2] <- 1
    label[ind4] <- 2

    return(list(X=X, Y=Y, label=label))
}

# This function generates a hex-cluster classification problem with three 
# classes. Cluster centers are at the six corners of a hexagon with each 
# two clusters across the center belonging to the same class, so it is 
# an generalized XOR problem and linearly inseperable. 
# input
#     var: variance of each Gaussian cluster
#     add.noise: whether to make it a superset label problem
# output
#     X: covariates
#     Y: labels as a matrix of binary indicators
#     label: labels as a vector 


hexClust <- function(var = 0.2, add.noise = FALSE)
{
    var <- 0.2
    n.train <- 600
    X <- matrix(rnorm(n.train * 2, mean=0, sd=var), ncol=2)
    Y <- matrix(0, nrow=nrow(X), ncol=3)
    label <- rep(0, n.train)

    sq3 <- 1.7320508
    center <- list()
    center[[1]] <- matrix(c(0, 2), ncol=2) * 0.5
    center[[2]] <- matrix(c(sq3, 1), ncol=2) * 0.5 
    center[[3]] <- matrix(c(sq3, -1), ncol=2) * 0.5 
    center[[4]] <- matrix(c(0, -2), ncol=2) * 0.5 
    center[[5]] <- matrix(c(-sq3, -1), ncol=2) * 0.5 
    center[[6]] <- matrix(c(-sq3, 1), ncol=2) * 0.5 

    # probability of not adding distractor label, adding distractor 
    # label with small index and adding distractor label with big index.
    prob <- c(1/3, 1/3, 1/3)
    labset <- 1:3 
    
    # cluster 1 belongs to class 1
    for (i.clust in 1:6)
    {
        # assign class for the current cluster
        cls <- (i.clust - 1) %% 3 + 1
        index <- ((i.clust - 1) * n.train / 6 + 1) : (i.clust * n.train / 6)
        X[index, ] <- X[index, ] + center[[i.clust]][rep(1, length(index)), ] 
        Y[index, cls] <- 1
        label[index] <- cls 

        if (add.noise)
        {
            rand.ind <- sample(c(1, 2, 3), length(index), replace=TRUE, prob=prob)
            sup.ind1 <- index[rand.ind == 1]
            sup.ind2 <- index[rand.ind == 2]

            small.distractor <- min(labset[labset != cls])
            big.distractor <- max(labset[labset != cls])

            Y[sup.ind1, small.distractor] <- 1
            Y[sup.ind2, big.distractor] <- 1
        }
    }
    
    return(list(X=X, Y=Y, label=label))
}

drawBoundary <- function(model, xlim, ylim)
{
    for (x1 in seq(xlim[1], xlim[2], by = 0.01)) 
    {
        for (x2 in seq(ylim[1], ylim[2], by = 0.01))
        {
            X <- matrix(c(x1, x2), nrow = 1)
            prob <- predict(model, X)
            prob <- sort(prob, decreasing = TRUE)
            diff <- abs(prob[1] - prob[2]) 
            if (abs(diff) < 0.015)
            {
                lines(x1, x2, type="p", pch=".", col="black")
            }
        }
    }
}

